// NimExtended.java

import ch.aplu.jgamegrid.*;
import java.awt.Color;
import java.awt.Font;
import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public class NimExtended extends GameGrid
  implements GGMouseListener, GGButtonListener
{
  private int nbMatches;
  private int nbTakenMatches;
  private GGBackground bg;
  private int row;
  private int[] igMatches;
  private final int dualMax = 4;
  private GGButton okBtn = new GGButton("sprites/ok.gif", true);
  private GGButton newBtn = new GGButton("sprites/new.gif", true);

  public NimExtended()
  {
    super(52, 44, 12, null, false);
    bg = getBg();
    igMatches = new int[44];
    addActor(okBtn, new Location(47, 3));
    okBtn.addButtonListener(this);
    addActor(newBtn, new Location(47, 3));
    newBtn.addButtonListener(this);
    init();
    addMouseListener(this, GGMouse.lPress);
    show();
  }

  public void init()
  {
    nbTakenMatches = 0;
    row = 0;
    bg.clear();
    for (int y = 4; y < 44; y = y + 7)
    {
      for (int i = 0; i < Math.random() * 15; i++)
      {
        addActor(new Match(), new Location(2 + 3 * i, y));
        igMatches[y] = igMatches[y] + 1;
        nbMatches++;
      }
    }
    okBtn.show();
    newBtn.hide();
    setTitle(nbMatches + " matches. Remove any amount of matches from one row.");
  }

  public boolean mouseEvent(GGMouse mouse)
  {
    Location loc = toLocationInGrid(mouse.getX(), mouse.getY());
    Actor actor = null;
    for (int y = -3; y < 4; y++)
    {
      actor = getOneActorAt(new Location(loc.x, loc.y + y), Match.class);
      if (actor != null)
        break;
    }
    if (actor != null)
    {
      if (row != 0 && actor.getY() != row)
        setTitle("You can only take matches from the same row");
      else
      {
        row = actor.getY();
        nbTakenMatches--;
        nbMatches--;
        igMatches[actor.getY()] = igMatches[actor.getY()] - 1;
        setTitle(nbMatches + " matches remaining. Click 'OK' to continue.");
        actor.removeSelf();  
        if (nbMatches == 0)
        {
          bg.setPaintColor(Color.red);
          bg.setFont(new Font("Arial", Font.BOLD, 96));
          bg.drawText("You win!", new Point(toPoint(new Location(3, 7))));
          okBtn.hide();
          newBtn.show();
          setTitle("You win, press 'New Game' to play again.");
        }
        refresh();
      }
    }
    return true;
  }

  public void buttonClicked(GGButton button)
  {
    if (nbMatches == 0)
      init();
    else
    {
      if (nbTakenMatches == 0)
        setTitle("You have to remove at least 1 match!");
      else
      {
        nbTakenMatches = 0;
        computerMove();
      }
    }
  }

  public void buttonPressed(GGButton button)
  {
  }

  public void buttonReleased(GGButton button)
  {
  }

  private void computerMove()
  {
    int row1 = 0;
    int nbToRemoveMatches = 0;
    // if optimal Strategy is not possible, do something random.
    if (!isUSituation(igMatches))
    {
      ArrayList<Actor> matches = getActors(Match.class);
      // from a random (not empty!) row
      Collections.shuffle(matches);
      row1 = matches.get(0).getY();
      // take a random amount
      nbToRemoveMatches = (int)((igMatches[row1] - 1) * Math.random() + 1);
      System.out.println("Player is winning");
    }
    else
    {
      System.out.println("Computer is winning! (yay)");
      // list for saving all possible solutions
      List<int[]> solutions = new ArrayList<int[]>();
      int[] tgMatches = new int[44];
      // Try all possible situations and add them to "solutions if they're good.
      for (int y = 4; y < 45; y = y + 7)
      {
        System.arraycopy(igMatches, 0, tgMatches, 0, 44);
        row1 = y;
        for (int i = 0; i < igMatches[y]; i++)
        {
          tgMatches = makeSituation(tgMatches, y);
          if (isUSituation(tgMatches) == false)
          {
            solutions.add((new int[]
              {
                y, i + 1
              }));
          }
        }
      }
      // choose a random solution
      Collections.shuffle(solutions);
      row1 = solutions.get(0)[0];
      nbToRemoveMatches = solutions.get(0)[1];
      System.out.println("Number of solutions: " + solutions.size());
    }

    System.out.println("Computer removes: Row: " + (((row1 - 4) / 7) + 1) + " Amount: " + nbToRemoveMatches);
    nbMatches = nbMatches - nbToRemoveMatches;
    ArrayList<Actor> matches = getActors(Match.class);
    Collections.shuffle(matches);
    int x = 0;
    while (nbToRemoveMatches > 0)
    {
      if (matches.get(x).getY() == row1)
      {
        matches.get(x).removeSelf();
        igMatches[row1] = igMatches[row1] - 1;
        nbToRemoveMatches--;
      }
      x++;
    }
    setTitle(nbMatches + " matches remaining. Your move now.");

    if (nbMatches == 0)
    {
      setTitle("Press 'new Game' to play again.");
      bg.setPaintColor(Color.red);
      bg.setFont(new Font("Arial", Font.BOLD, 90));
      bg.drawText("You lost!", new Point(toPoint(new Location(3,
        7))));
      okBtn.hide();
      newBtn.show();
    }
    refresh();
    row = 0; // Spieler darf neue "Ziehreihe" bestimmen
  }

   // for debugging only

  private String toString(int[] k)
  {
    String output = "";
    for (int i = 0; i < k.length; i++)
      output = (output + k[i] + ", ");
    return output;
  }
  /*
   * removes a match from a situation given the:
   * "sit", the original situation
   * "row" the row where a match should be removed from
   */

  private int[] makeSituation(int[] sit, int row)
  {
    sit[row] = sit[row] - 1;
    return sit;
  }
  // Check if its a U-Situation

  private Boolean isUSituation(int[] sit)
  {
    int[] allDuals = new int[dualMax];
    int[] oneDual = new int[dualMax];
    for (int y = 4; y < 45; y = y + 7)
    {
      oneDual = toDual(sit[y]);
      for (int i = 0; i < allDuals.length; i++)
      {
        allDuals[i] = allDuals[i] + oneDual[i];
      }
    }
    for (int i = 0; i < allDuals.length; i++)
    {
      if (allDuals[i] % 2 == 1)
        return true;
    }
    return false;
  }

  /*
   * gives an integer back in binary (as array of integer) only works for
   * input < 16, or arraylenght "dualMax" must be changed
   */
  private int[] toDual(int input)
  {
    int[] output = new int[dualMax]; // bei hÃ?Â¶chstens 15 StreichhÃ?Â¶lzern
    // reichen 4 dualstellen
    int x = 0;
    while (input != 0)
    {
      output[x] = input % 2;
      input = input / 2;
      x++;
    }
    return output;
  }

  public static void main(String[] args)
  {
    new NimExtended();
  }
}

// ---------------- class Match --------------------------
class Match extends Actor
{
  public Match()
  {
    super("sprites/match.gif");
  }
}
